///
/// Custom FM radio with serial control and RDS monitor.
/// Copyright Jonathan Cohen, 2020-05-29 to 2023-08-26.
///
/// Based on code by Matthias Hertel.
///
/// Changelog:
///     2020-10-22 Added use of internal EEPROM to store selected frequency in non-volatile memory for unattended startup.
///     2023-08-26 Checked and tidy of code to run on V2 of the Arduino IDE.
//
/// Radio reception tested with RDA5807M and SI4703.
/// SI4703 had better selectivity and sensitivity but needs an extra reset wire.
/// Don't use the RDA5807M as there may be software differences.
///
/// Serial interface is 115200 baud.
///
///
/// Arduino port | SI4703 signal | RDA5807M signal
/// :----------: | :-----------: | :-------------:
/// GND (black)  | GND           | GND   
/// 3.3V (red)   | VCC           | VCC
/// 5V (red)     | -             | -
/// A5 (yellow)  | SCLK          | SCLK
/// A4 (blue)    | SDIO          | SDIO
/// D2           | RST           | -


#include <Wire.h>
#include <radio.h>
#include <RDA5807M.h>
#include <SI4703.h>
#include <SI4705.h>
#include <TEA5767.h>
#include <RDSParser.h>
#include <EEPROM.h>         // Added 2020-10-22.


// Some presets. Can't be changed by the user.
RADIO_FREQ preset[] = {
  10510,
  10110,
  10260,
  10680,
  10790
 };


int i_sidx = 0;       // < Start at Station with index=0


// EEPROM Frequency storage
long efreq1 = EEPROM.read(0);      // Frequency byte 0 in EEPROM 0.
long efreq2 = EEPROM.read(1);      // Frequency byte 1 in EEPROM 1.
long eefreq = (efreq1 *256) + efreq2;
// EEPROM Volume storage
long evol = EEPROM.read(2);       // Volume in EEPROM 2.


char rdsServiceName[8 + 2]; ///< String with the actual RDS Service name. Some senders put rotating text in here too.
char rdsText[64 + 2];       ///< String with the actual RDS text.
char rdsTime[6];            ///< String with the actual time from RDS as hh:mm



/// The radio object has to be defined by using the class corresponding to the used chip.
/// by uncommenting the right radio object definition.
//
// RADIO radio;       ///< Create an instance of a non functional radio.
// RDA5807M radio;    ///< Create an instance of a RDA5807 chip radio
SI4703 radio;    ///< Create an instance of a SI4703 chip radio.
// SI4705   radio;    ///< Create an instance of a SI4705 chip radio.
// TEA5767  radio;    ///< Create an instance of a TEA5767 chip radio.


/// get a RDS parser
RDSParser rds;


/// State definition for this radio implementation.
enum RADIO_STATE {
  STATE_NONE = 0,
  STATE_PARSECOMMAND, ///< waiting for a new command character.
  STATE_PARSEINT,     ///< waiting for digits for the parameter.
  STATE_EXEC,         ///< executing the command.

  STATE_RDS,     ///< display RDS information
  STATE_VOL,
  STATE_PRESET,
  STATE_FREQ,
  STATE_MONO,
  STATE_SMUTE

};

RADIO_STATE state; ///< The state variable is used for parsing input characters.
RADIO_STATE rot_state;


// - - - - - - - - - - - - - - - - - - - - - - - - - -

void loopRadio(unsigned long now);



/// Update the Frequency on the serial display.
void DisplayFrequency(RADIO_FREQ f)
{
  char s[12];
  radio.formatFrequency(s, sizeof(s));
  Serial.print("FR: "); Serial.println(s);
} // DisplayFrequency()




// This function will be called by the RDS module when a rds service name was received.
/// The text be displayed on the LCD and written to the serial port
/// and will be stored for the web interface.
void DisplayServiceName(char *name)
{
  Serial.print("PS: ");
  Serial.println(name);
  
//  DEBUG_VAL("RDS", name);
//  strncpy(rdsServiceName, name, sizeof(rdsServiceName));

//  if (rot_state == STATE_RDS) {
  //  lcd.setCursor(0, 1);
 //   lcd.print(name);
 // }
} // DisplayServiceName()


/// This function will be called by the RDS module when a rds text message was received.
/// The text will not displayed on the LCD but written to the serial port
/// and will be stored for the web interface.
void DisplayText(char *text)
{
  Serial.print("RT: ");
  Serial.println(text);
//  digitalWrite(LED_BUILTIN, HIGH);       //test
//  delay(50);
//  digitalWrite(LED_BUILTIN, LOW);       //test

  //DEBUG_VAL("RDS-text", text);
  //strncpy(rdsText, text, sizeof(rdsText));
} // DisplayText()


/// This function will be called by the RDS module when a rds time message was received.
/// The time will not displayed on the LCD but written to the serial port.
void DisplayTime(uint8_t hour, uint8_t minute) {
  rdsTime[0] = '0' + (hour / 10);
  rdsTime[1] = '0' + (hour % 10);
  rdsTime[2] = ':';
  rdsTime[3] = '0' + (minute / 10);
  rdsTime[4] = '0' + (minute % 10);
  //rdsTime[5] = NUL;
    Serial.print("CT: ");
  Serial.println(rdsTime);
//  DEBUG_VAL("RDS-time", rdsTime);
} // DisplayTime()




// - - - - - - - - - - - - - - - - - - - - - - - - - -



/// Retrieve RDS data from the radio chip and forward to the RDS decoder library
///void RDS_process(uint16_t block1, uint16_t block2, uint16_t block3, uint16_t block4) {
///static uint16_t b1;

///  if (b1 != block1)
//    Serial.print("PI: ");       // Not actually sure if it always is :-)
///    Serial.println(block1,HEX);
//    Serial.print(block2,HEX);
//    Serial.print(block3,HEX);
//    Serial.println(block4,HEX);
///    rds.processData(block1, block2, block3, block4);
///    b1 = block1;
///}

void RDS_process(uint16_t block1, uint16_t block2, uint16_t block3, uint16_t block4) {
  rds.processData(block1, block2, block3, block4);
}


/// Execute a command identified by a character and an optional number.
/// See the "?" command for available commands.
/// \param cmd The command character.
/// \param value An optional parameter for the command.
void runSerialCommand(char cmd, int16_t value)
{
  if (cmd == '?') {
    //
//    radio.setFrequency(eefreq);     // Before display, set the radio frequency to the value stored in the EEPROM so this happens at boot!
//    radio.setVolume(evol);          // Before display, set the volume level to the value stored in the EEPROM so this happens at boot.
    //
    Serial.println("\n");
    Serial.println("TERMINAL COMMANDS");
    Serial.println("-----------------");
    Serial.println("? - Display this text");
//    Serial.println("+ increase volume");
//    Serial.println("- decrease volume");
//    Serial.println("> next preset");
//    Serial.println("< previous preset");
//    Serial.println(". tune up   : scan up to next sender");
//    Serial.println(", tune down ; scan down to next sender");
    Serial.println("fnnnnn - Select frequency and store in EEPROM");
    Serial.println("vnn - Set volume 0-15 and store in EEPROM");
    Serial.println("sn - 0=Stereo, 1=Force Mono (normal=0)");
    Serial.println("b - 0=Bass Boost Off, 1=Bass Boost On (normal=0)");
    Serial.println("i - Display station status");
    Serial.println("u - Mute / Unmute audio");
    Serial.println("x - Display radio chip diagnostics");
    Serial.println("! - Init radio chip");
    Serial.println("");
    Serial.print("Forced Mono (0=Stereo, 1=Mono): "); Serial.println(radio.getMono());
    Serial.print("Bass Boost (0=off, 1=on): "); Serial.println(radio.getBassBoost());
    Serial.print("EEPROM Freq: "); Serial.println(eefreq);
    Serial.print("EEPROM Volume: "); Serial.println(evol);
    Serial.println();
  }

  // ----- control the volume and audio output -----

  else if (cmd == '+') {
    // increase volume
    int v = radio.getVolume();
    if (v < 15) radio.setVolume(++v);
  } else if (cmd == '-') {
    // decrease volume
    int v = radio.getVolume();
    if (v > 0) radio.setVolume(--v);
  }

  else if (cmd == 'u') {
    // toggle mute mode
    //digitalWrite(LED_BUILTIN, HIGH);       //test
    radio.setMute(!radio.getMute());
  }

  // set stereo mode
  else if (cmd == 's') { 
    //radio.setMono(!radio.getMono()); 
    if (value == 0) { radio.setMono(0); }
    if (value == 1) { radio.setMono(1); }
  }

  // toggle bass boost
  else if (cmd == 'b') { radio.setBassBoost(!radio.getBassBoost()); }


  else if (cmd == 'v') {
    // Set and store volume
    if (value < 16) {
      radio.setVolume(value);
      Serial.print("Storing volume in EEPROM: "); Serial.println(value);
      EEPROM.write(2, value);     // location, value
    }
  }

  
  // ----- control the frequency -----

  else if (cmd == '>') {
    // next preset
    if (i_sidx < (sizeof(preset) / sizeof(RADIO_FREQ)) - 1) {
      i_sidx++; radio.setFrequency(preset[i_sidx]);
    } // if
  } else if (cmd == '<') {
    // previous preset
    if (i_sidx > 0) {
      i_sidx--;
      radio.setFrequency(preset[i_sidx]);
    } // if

  } else if (cmd == 'f') { 
    // Change frequency:
    radio.setFrequency(value); 
    // Store the set frequency in EEPROM:
    Serial.print("Storing frequency in EEPROM: "); Serial.println(value);
    efreq1=int(value/256);
    efreq2=int(value - (efreq1*256));
    // Store
//    Serial.print("0: "); Serial.println(efreq1);
//    Serial.print("1: "); Serial.println(efreq2);
    EEPROM.write(0, efreq1);     // location, value
    EEPROM.write(1, efreq2);     // location, value
  }

  else if (cmd == '.') { radio.seekUp(false); } else if (cmd == ':') { radio.seekUp(true); } else if (cmd == ',') { radio.seekDown(false); } else if (cmd == ';') { radio.seekDown(true); }


  // not in help:
  else if (cmd == '!') {
    if (value == 0) radio.term();
    if (value == 1) radio.init();

  } else if (cmd == 'i') {
    char s[12];
    radio.formatFrequency(s, sizeof(s));
    Serial.print("Station:"); Serial.println(s);
    Serial.println("Radio:"); radio.debugRadioInfo();
    Serial.println("Audio:"); radio.debugAudioInfo();
  // radio.debugRadioInfo returns:
  // RDS TUNED  MONO   RSSI: ?  SNR: ?
  // --- ----- STEREO  RSSI: ?  SNR: ?
  // radio.debugAudioInfo returns:
  // MUTE SOFTMUTE BASS
  // ---- -------- ---- 


  //
  } // info

  else if (cmd == 'x') {
    radio.debugStatus(); // print chip specific data.
  }
} // runSerialCommand()




/// Check once new radio data.
void loopRadio(unsigned long now) {
  // check for RDS data
  radio.checkRDS();
} // loopRadio()





/// Setup a FM only radio configuration with I/O for commands and debugging on the Serial port.
void setup() {
  pinMode(LED_BUILTIN, OUTPUT);   // Enable the L LED as an output.
  // open the Serial port
  Serial.begin(115200);
  Serial.print("\nReadyformed FM receiver and RDS analyser. www.readyformed.com firmware v1.02\n\n");
  delay(500);

  // Initialize the Radio 
  radio.init();

  // Enable information to the Serial port
  radio.debugEnable();

  //radio.setBandFrequency(RADIO_BAND_FM, preset[i_sidx]); // 5. preset.
  radio.setBandFrequency(RADIO_BAND_FM, eefreq);

  // delay(100);

  radio.setMono(false);
  radio.setMute(false);
  // radio.debugRegisters();
  radio.setVolume(evol);      (8);   // was 12

  Serial.write('>');

  state = STATE_PARSECOMMAND;

  // setup the information chain for RDS data.
  radio.attachReceiveRDS(RDS_process);
  rds.attachServicenNameCallback(DisplayServiceName);
  //new by JC
  rds.attachTextCallback(DisplayText);
  rds.attachTimeCallback(DisplayTime);
  

  runSerialCommand('?', 0);
} // Setup


/// Constantly check for serial input commands and trigger command execution.
void loop() {




  int newPos;
  unsigned long now = millis();
  static unsigned long nextFreqTime = 0;
  static unsigned long nextRadioInfoTime = 0;


//    loopSerial(now);     /// Check for serial input commands and trigger command execution.
  loopRadio(now);      /// Check for new radio data.


  // some internal static values for parsing the input
  static char command;
  static int16_t value;
  static RADIO_FREQ lastf = 0;
  RADIO_FREQ f = 0;

  char c;
  if (Serial.available() > 0) {
    // read the next char from input.
    c = Serial.peek();

    if ((state == STATE_PARSECOMMAND) && (c < 0x20)) {
      // ignore unprintable chars
      Serial.read();

    } else if (state == STATE_PARSECOMMAND) {
      // read a command.
      command = Serial.read();
      state = STATE_PARSEINT;

    } else if (state == STATE_PARSEINT) {
      if ((c >= '0') && (c <= '9')) {
        // build up the value.
        c = Serial.read();
        value = (value * 10) + (c - '0');
      } else {
        // not a value -> execute
        runSerialCommand(command, value);
        command = ' ';
        state = STATE_PARSECOMMAND;
        value = 0;
      } // if
    } // if
  } // if


  // check for RDS data
  radio.checkRDS();

  // update the display from time to time
  if (now > nextFreqTime) {
    f = radio.getFrequency();
    if (f != lastf) {
      // print current tuned frequency
      DisplayFrequency(f);
      lastf = f;
    } // if
    nextFreqTime = now + 400;
  } // if  

} // loop



// End.
